package me.libraryaddict.disguise;

import lombok.Getter;
import lombok.SneakyThrows;
import me.libraryaddict.disguise.commands.LibsDisguisesCommand;
import me.libraryaddict.disguise.commands.disguise.DisguiseCommand;
import me.libraryaddict.disguise.commands.disguise.DisguiseEntityCommand;
import me.libraryaddict.disguise.commands.disguise.DisguisePlayerCommand;
import me.libraryaddict.disguise.commands.disguise.DisguiseRadiusCommand;
import me.libraryaddict.disguise.commands.modify.DisguiseModifyCommand;
import me.libraryaddict.disguise.commands.modify.DisguiseModifyEntityCommand;
import me.libraryaddict.disguise.commands.modify.DisguiseModifyPlayerCommand;
import me.libraryaddict.disguise.commands.modify.DisguiseModifyRadiusCommand;
import me.libraryaddict.disguise.commands.undisguise.UndisguiseCommand;
import me.libraryaddict.disguise.commands.undisguise.UndisguiseEntityCommand;
import me.libraryaddict.disguise.commands.undisguise.UndisguisePlayerCommand;
import me.libraryaddict.disguise.commands.undisguise.UndisguiseRadiusCommand;
import me.libraryaddict.disguise.commands.utils.CopyDisguiseCommand;
import me.libraryaddict.disguise.commands.utils.DisguiseCloneCommand;
import me.libraryaddict.disguise.commands.utils.DisguiseHelpCommand;
import me.libraryaddict.disguise.commands.utils.DisguiseViewBarCommand;
import me.libraryaddict.disguise.commands.utils.DisguiseViewSelfCommand;
import me.libraryaddict.disguise.commands.utils.GrabHeadCommand;
import me.libraryaddict.disguise.commands.utils.GrabSkinCommand;
import me.libraryaddict.disguise.commands.utils.SaveDisguiseCommand;
import me.libraryaddict.disguise.utilities.DisguiseUtilities;
import me.libraryaddict.disguise.utilities.LibsPremium;
import me.libraryaddict.disguise.utilities.config.DisguiseCommandConfig;
import me.libraryaddict.disguise.utilities.listeners.DisguiseListener;
import me.libraryaddict.disguise.utilities.listeners.DisguiseListener1_18;
import me.libraryaddict.disguise.utilities.listeners.PaperDisguiseListener;
import me.libraryaddict.disguise.utilities.listeners.PlayerSkinHandler;
import me.libraryaddict.disguise.utilities.metrics.MetricsInitalizer;
import me.libraryaddict.disguise.utilities.packets.PacketsManager;
import me.libraryaddict.disguise.utilities.params.ParamInfoManager;
import me.libraryaddict.disguise.utilities.parser.DisguiseParser;
import me.libraryaddict.disguise.utilities.reflection.ClassMappings;
import me.libraryaddict.disguise.utilities.reflection.NmsVersion;
import me.libraryaddict.disguise.utilities.reflection.ReflectionManager;
import me.libraryaddict.disguise.utilities.sounds.SoundManager;
import me.libraryaddict.disguise.utilities.updates.UpdateChecker;
import me.libraryaddict.disguise.utilities.watchers.CompileMethods;
import org.apache.commons.lang.StringUtils;
import org.bukkit.Bukkit;
import org.bukkit.command.Command;
import org.bukkit.command.CommandExecutor;
import org.bukkit.command.CommandMap;
import org.bukkit.command.PluginCommand;
import org.bukkit.command.TabCompleter;
import org.bukkit.configuration.file.YamlConfiguration;
import org.bukkit.plugin.Plugin;
import org.bukkit.plugin.java.JavaPlugin;
import org.bukkit.scheduler.BukkitRunnable;
import org.jetbrains.annotations.NotNull;

import java.io.File;
import java.util.Arrays;
import java.util.Iterator;
import java.util.Map;
import java.util.stream.Collectors;

public class LibsDisguises extends JavaPlugin {
    private static LibsDisguises instance;
    private DisguiseListener listener;
    private String buildNumber;
    @Getter
    private String buildDate;
    @Getter
    private boolean reloaded;
    @Getter
    private final UpdateChecker updateChecker = new UpdateChecker();
    @Getter
    private PlayerSkinHandler skinHandler;
    private DisguiseCommandConfig commandConfig;

    @Override
    public void onLoad() {
        try {
            if (instance != null || !Bukkit.getServer().getWorlds().isEmpty() || !Bukkit.getOnlinePlayers().isEmpty()) {
                reloaded = true;
                getLogger().severe("Server was reloaded! Please do not report any bugs! This plugin can't handle " + "reloads gracefully!");
            }

            instance = this;

            Plugin plugin = Bukkit.getPluginManager().getPlugin("ProtocolLib");

            if (plugin == null || DisguiseUtilities.isProtocolLibOutdated()) {
                if (DisguiseConfig.isNeverUpdateProtocolLib()) {
                    getLogger().warning(
                        "Defined in plugins/LibsDisguises/configs/sanity.yml, you have requested that Lib's Disguises never updates or " +
                            "installs ProtocolLib. " + "Please do not report any issues with this plugin.");
                } else {
                    String reason;

                    if (plugin == null) {
                        reason = "ProtocolLib not installed";
                    } else if (!plugin.isEnabled()) {
                        reason = "ProtocolLib is not enabled";
                    } else if (DisguiseUtilities.isProtocolLibOutdated()) {
                        reason = "ProtocolLib is outdated";
                    } else {
                        reason = "Lib's Disguises couldn't process ProtocolLib properly";
                    }

                    getLogger().warning(
                        "An issue occured when trying to load ProtocolLib: " + reason + ". Lib's Disguises will attempt to update it.");

                    try {
                        File dest = DisguiseUtilities.updateProtocolLib();

                        if (plugin == null) {
                            getLogger().info("ProtocolLib downloaded and stuck in plugins folder! Now trying to load it!");
                            plugin = Bukkit.getPluginManager().loadPlugin(dest);
                            plugin.onLoad();

                            Bukkit.getPluginManager().enablePlugin(plugin);
                        } else {
                            getLogger().severe("Please restart the server to complete the ProtocolLib update!");
                        }
                    } catch (Exception e) {
                        getLogger().severe("Looks like ProtocolLib's site may be down! Try download it manually from https://ci.dmulloy2" +
                            ".net/job/ProtocolLib/");
                        e.printStackTrace();
                    }
                }
            }

            try {
                Class cl = Class.forName("org.bukkit.Server$Spigot");
            } catch (ClassNotFoundException e) {
                getLogger().severe("Oh dear, you seem to be using CraftBukkit. Please use Spigot or Paper instead! This " +
                    "plugin will continue to load, but it will look like a mugging victim");
            }

            // Note that Forge/Hybrid is unsupported
            // Previously was supported but changes in both LD and nms has made it unsustainable
            // The time and complexibility it would take is far too high for the amount of usage it would receive.
            // Any extra income from sales would never cover the time spent implementing this.
            if (Bukkit.getVersion().toLowerCase().matches(".*(arclight|mohist|magma).*")) {
                getLogger().severe(
                    "Lib's Disguises will not work correctly on Forge/Hybrid servers, due to complexibility and time commitment there is " +
                        "no support for this. Do not report any issues to Lib's Disguises.");
            }

            commandConfig = new DisguiseCommandConfig();

            if (!reloaded) {
                commandConfig.load();
            }
        } catch (Throwable throwable) {
            try {
                if (isNumberedBuild() && DisguiseConfig.isAutoUpdate()) {
                    getUpdateChecker().doUpdate();
                }
            } catch (Throwable t) {
                getLogger().severe("Failed to even do a forced update");
            }

            throw throwable;
        }
    }

    @SneakyThrows
    @Override
    public void onEnable() {
        try {
            if (reloaded) {
                getLogger().severe("Server was reloaded! Please do not report any bugs! This plugin can't handle " + "reloads gracefully!");
            }

            if (Bukkit.getVersion().contains("(MC: 1.17)")) {
                getLogger().severe("Please update from MC 1.17! You should be using 1.17.1!");
            }

            if (Bukkit.getVersion().contains("(MC: 1.18)") || Bukkit.getVersion().contains("(MC: 1.18.1)")) {
                getLogger().severe(
                    "Please update from MC 1.18 and MC 1.18.1! You should be using 1.18.2! Support will eventually be dropped for your " +
                        "specific version!");
            }

            if (Bukkit.getVersion().contains("(MC: 1.19)")) {
                getLogger().severe(
                    "Please update from MC 1.19.0! You should be using at least 1.19.3! 1.19.1 is the lowest supported 1.19 version!");
            }

            try {
                Class cl = Class.forName("org.bukkit.Server$Spigot");
            } catch (ClassNotFoundException e) {
                getLogger().severe("Oh dear, you seem to be using CraftBukkit. Please use Spigot or Paper instead! This " +
                    "plugin will continue to load, but it will look like a mugging victim");
            }

            File disguiseFile = new File(getDataFolder(), "configs/disguises.yml");

            if (!disguiseFile.exists()) {
                disguiseFile.getParentFile().mkdirs();

                File oldFile = new File(getDataFolder(), "disguises.yml");

                if (oldFile.exists()) {
                    oldFile.renameTo(disguiseFile);
                } else {
                    saveResource("configs/disguises.yml", false);
                }
            }

            YamlConfiguration pluginYml = ReflectionManager.getPluginYAML(getFile());
            buildNumber = StringUtils.stripToNull(pluginYml.getString("build-number"));
            buildDate = StringUtils.stripToNull(pluginYml.getString("build-date"));

            int fileCount = ReflectionManager.getJarFileCount(getFile(), CompileMethods.ignoredDirectories());
            int expected = pluginYml.getInt("file-count", fileCount);

            if (fileCount != expected) {
                getLogger().severe("Lib's Disguises may be infected with malware, there were " + (fileCount - expected) +
                    " unknown files injected into the jar. Please redownload from a trusted source such as SpigotMC. If this warning " +
                    "shows even after updating, try https://www.spigotmc.org/resources/spigot-anti-malware.64982/ otherwise if that " +
                    "doesn't help then please contact libraryaddict on discord https://discord.gg/J5XRqYX");
            }

            getLogger().info("File Name: " + getFile().getName());

            String nmsPackageName = ReflectionManager.getNmsPackage();

            if (nmsPackageName.isEmpty()) {
                nmsPackageName = "{Not package relocated}";
            }

            getLogger().info(
                "Discovered nms version: (Package: " + nmsPackageName + ") (LD: " + ReflectionManager.getVersion() + ") (MC: " +
                    ReflectionManager.getMinecraftVersion() + ")");

            getLogger().info("Jenkins Build: " + (isNumberedBuild() ? "#" : "") + getBuildNo());

            getLogger().info("Build Date: " + buildDate);

            DisguiseConfig.loadInternalConfig();

            LibsPremium.check(getDescription().getVersion(), getFile());

            if (!LibsPremium.isPremium()) {
                getLogger().info(
                    "You are running the free version, commands limited to non-players and operators. (Console," + " Command " +
                        "Blocks, Admins)");
            }

            // Add a message so people are more aware
            if (!DisguiseConfig.isTallSelfDisguises()) {
                DisguiseUtilities.getLogger().info(
                    "Config 'TallSelfDisguises' is set to 'false', LD will hide oversized disguises from self disguise. https://www" +
                        ".spigotmc.org/wiki/lib-s-disguises-faq/#tall-disguises-self-disguises");
            }

            if (ReflectionManager.getVersion() == null) {
                getLogger().severe("You're using the wrong version of Lib's Disguises for your server! This is " + "intended for " +
                    StringUtils.join(Arrays.stream(NmsVersion.values()).filter(v -> v != NmsVersion.UNSUPPORTED)
                        .map(v -> String.join(", ", v.getSupportedVersions())).collect(Collectors.toList()), ", ") + "!");
                getPluginLoader().disablePlugin(this);
                return;
            }

            if (!DisguiseConfig.isNeverUpdateProtocolLib() && DisguiseUtilities.isProtocolLibOutdated()) {
                String requiredProtocolLib = StringUtils.join(DisguiseUtilities.getProtocolLibRequiredVersion(), " or build #");
                String version = Bukkit.getPluginManager().getPlugin("ProtocolLib").getDescription().getVersion();

                BukkitRunnable runnable = createProtocolLibOutdatedRunnable(version, requiredProtocolLib);
                runnable.run();
                runnable.runTaskTimer(this, 20, 10 * 60 * 20);
            }

            // If this is a release build, even if jenkins build..
            if (isReleaseBuild()) {
                // If downloaded from spigot, forcibly set release build to true
                if (LibsPremium.getUserID().matches("\\d+")) {
                    DisguiseConfig.setUsingReleaseBuilds(true);
                }
                // Otherwise leave it untouched as they might've just happened to hit a dev build, which is a release build
            } else {
                DisguiseConfig.setUsingReleaseBuilds(false);
            }

            ClassMappings.loadMappingsCache(getDataFolder());

            ReflectionManager.init();

            PacketsManager.init();
            DisguiseUtilities.init();

            new SoundManager().load();

            ReflectionManager.registerValues();
            DisguiseParser.createDefaultMethods();

            ParamInfoManager.getParamInfoSoundGroup().recalculate();

            DisguiseConfig.loadConfig();

            PacketsManager.addPacketListeners();

            listener = new DisguiseListener(this);
            skinHandler = new PlayerSkinHandler();

            Bukkit.getPluginManager().registerEvents(getSkinHandler(), LibsDisguises.getInstance());

            if (NmsVersion.v1_18.isSupported()) {
                Bukkit.getPluginManager().registerEvents(new DisguiseListener1_18(), this);
            }

            if (DisguiseUtilities.isRunningPaper()) {
                Bukkit.getPluginManager().registerEvents(new PaperDisguiseListener(), this);
            }

            registerCommand("libsdisguises", new LibsDisguisesCommand());
            registerCommand("disguise", new DisguiseCommand());
            registerCommand("undisguise", new UndisguiseCommand());
            registerCommand("disguiseplayer", new DisguisePlayerCommand());
            registerCommand("undisguiseplayer", new UndisguisePlayerCommand());
            registerCommand("undisguiseentity", new UndisguiseEntityCommand());
            registerCommand("disguiseentity", new DisguiseEntityCommand());
            registerCommand("disguiseradius", new DisguiseRadiusCommand());
            registerCommand("undisguiseradius", new UndisguiseRadiusCommand());
            registerCommand("disguisehelp", new DisguiseHelpCommand());
            registerCommand("disguiseclone", new DisguiseCloneCommand());
            registerCommand("disguiseviewself", new DisguiseViewSelfCommand());
            registerCommand("disguiseviewbar", new DisguiseViewBarCommand());
            registerCommand("disguisemodify", new DisguiseModifyCommand());
            registerCommand("disguisemodifyentity", new DisguiseModifyEntityCommand());
            registerCommand("disguisemodifyplayer", new DisguiseModifyPlayerCommand());
            registerCommand("disguisemodifyradius", new DisguiseModifyRadiusCommand());
            registerCommand("copydisguise", new CopyDisguiseCommand());
            registerCommand("grabskin", new GrabSkinCommand());
            registerCommand("savedisguise", new SaveDisguiseCommand());
            registerCommand("grabhead", new GrabHeadCommand());

            unregisterCommands(false);

            new MetricsInitalizer();
        } catch (Throwable throwable) {
            try {
                if (isNumberedBuild() && DisguiseConfig.isAutoUpdate()) {
                    getUpdateChecker().doUpdate();
                }
            } catch (Throwable t) {
                getLogger().severe("Failed to even do a forced update");
            }

            throw throwable;
        }
    }

    @NotNull
    private BukkitRunnable createProtocolLibOutdatedRunnable(String version, String requiredProtocolLib) {
        return new BukkitRunnable() {
            private int timesRun;

            @Override
            public void run() {
                getLogger().severe("!! May I have your attention please !!");

                if (DisguiseUtilities.isProtocollibUpdateDownloaded()) {
                    getLogger().severe(
                        "An update for ProtocolLib has been downloaded and will be installed when the server restarts. When possible, " +
                            "please restart " + "the server. Lib's Disguises may not work correctly until you do so.");
                } else {
                    getLogger().severe(
                        "Update your ProtocolLib! You are running " + version + " but the minimum version you should be on is " +
                            requiredProtocolLib + "!");
                    getLogger().severe(
                        "https://ci.dmulloy2.net/job/ProtocolLib/lastSuccessfulBuild/artifact/target" + "/ProtocolLib" + ".jar");
                    getLogger().severe("Or! Use /ld protocollib - To update to the latest development build");
                }

                if (timesRun++ > 0) {
                    getLogger().severe("This message is on repeat due to the sheer number of people who don't see this.");
                }

                getLogger().severe("!! May I have your attention please !!");
            }
        };
    }

    public void unregisterCommands(boolean force) {
        CommandMap map = ReflectionManager.getCommandMap();
        Map<String, Command> commands = ReflectionManager.getCommands(map);

        for (String command : getDescription().getCommands().keySet()) {
            PluginCommand cmd = getCommand("libsdisguises:" + command);

            if (cmd == null || (cmd.getExecutor() != this && !force)) {
                continue;
            }

            if (cmd.getPermission() != null && cmd.getPermission().startsWith("libsdisguises.seecmd")) {
                Bukkit.getPluginManager().removePermission(cmd.getPermission());
            }

            Iterator<Map.Entry<String, Command>> itel = commands.entrySet().iterator();

            while (itel.hasNext()) {
                Map.Entry<String, Command> entry = itel.next();

                if (entry.getValue() != cmd) {
                    continue;
                }

                itel.remove();
            }
        }
    }

    @Override
    public @NotNull File getFile() {
        return super.getFile();
    }

    @Override
    public void onDisable() {
        DisguiseUtilities.saveDisguises();
        ClassMappings.saveMappingsCache(getDataFolder());

        reloaded = true;
    }

    public boolean isReleaseBuild() {
        return !getDescription().getVersion().contains("-SNAPSHOT");
    }

    public String getBuildNo() {
        return buildNumber;
    }

    public int getBuildNumber() {
        return isNumberedBuild() ? Integer.parseInt(getBuildNo()) : 0;
    }

    public boolean isNumberedBuild() {
        return getBuildNo() != null && getBuildNo().matches("\\d+");
    }

    private void registerCommand(String commandName, CommandExecutor executioner) {
        String name = commandConfig.getCommand(commandName);

        if (name == null) {
            return;
        }

        PluginCommand command = getCommand("libsdisguises:" + name);

        if (command == null) {
            return;
        }

        command.setExecutor(executioner);

        if (executioner instanceof TabCompleter) {
            command.setTabCompleter((TabCompleter) executioner);
        }
    }

    /**
     * Reloads the config with new config options.
     */
    @Deprecated
    public void reload() {
        DisguiseConfig.loadConfig();
    }

    public DisguiseListener getListener() {
        return listener;
    }

    /**
     * External APIs shouldn't actually need this instance. DisguiseAPI should be enough to handle most cases.
     *
     * @return The instance of this plugin
     */
    public static LibsDisguises getInstance() {
        return instance;
    }
}
